import logging
import subprocess
import typing as t
from contextlib import AbstractContextManager
from dataclasses import dataclass
from os import path
from types import TracebackType

import psycopg2
import psycopg2.extensions
import requests
import unittest

from clazzes.osgitools.utils import retry_operation_simple
from clazzes.http.xmlrpc import LoginInfo, XmlRpcTargetError, JSON_RPC_LOGIN_REQUIRED, JSON_RPC_SERVER_ERROR_0, Bean

log = logging.getLogger(__name__)


@dataclass
class DockerComposeConfig:
    compose_directory: str
    # If None use the docker-compose defaults.
    compose_files: t.Optional[t.Sequence[str]] = None
    # If None do not save logs.
    log_file_path: t.Optional[str] = None
    run_args: t.Union[t.Sequence[str], str, None] = None


class DockerCompose(AbstractContextManager["DockerCompose"]):
    config: DockerComposeConfig

    def __init__(self, config: DockerComposeConfig):
        self.config = config

    def __enter__(self) -> "DockerCompose":
        compose_args = ["docker-compose"]

        if self.config.compose_files:
            for compose_file in self.config.compose_files:
                compose_args.append("-f")
                compose_args.append(compose_file)

        compose_args.extend(["up", "-d"])

        if self.config.run_args:
            if isinstance(self.config.run_args, str):
                compose_args.append(self.config.run_args)
            else:
                compose_args.extend(self.config.run_args)

        try:
            subprocess.check_call(compose_args, cwd=self.config.compose_directory)
        except:
            self.close()
            raise
        return self

    def close(self) -> None:
        if self.config.log_file_path is not None:
            try:
                stdout_file = self.config.log_file_path+"-stdout.log"
                stderr_file = self.config.log_file_path+"-stderr.log"
                with open(stdout_file,"wb") as stdout, open(stderr_file,"wb") as stderr:
                    subprocess.check_call(
                        ["docker-compose", "logs"],
                        cwd=self.config.compose_directory,
                        stdout=stdout,
                        stderr=stderr,
                    )
            except (KeyboardInterrupt,SystemExit):
                raise
            except:
                log.warn("Failed to write docker-compose logs", exc_info=True)

        try:
            subprocess.check_call(
                ["docker-compose", "down"], cwd=self.config.compose_directory
            )
        except (KeyboardInterrupt,SystemExit):
            raise
        except:
            log.warn("Failed to destroy docker-compose", exc_info=True)

    def __exit__(
        self,
        __exc_type: t.Optional[t.Type[BaseException]],
        __exc_value: t.Optional[BaseException],
        __traceback: t.Optional[TracebackType],
    ) -> t.Optional[bool]:
        self.close()
        return None


@dataclass
class OSGIRunnerConfig:
    osgi_dir: str
    run_args: t.Union[t.Sequence[str], str, None] = None


class OSGIRunner(AbstractContextManager["OSGIRunner"]):
    config: OSGIRunnerConfig
    osgi_runner: t.Optional[subprocess.Popen[bytes]]

    def __init__(self, config: OSGIRunnerConfig):
        self.config = config
        self.osgi_runner = None

    def __enter__(self) -> "OSGIRunner":
        popen_args = [path.join(self.config.osgi_dir, "run.sh")]

        if self.config.run_args:
            if isinstance(self.config.run_args, str):
                popen_args.append(self.config.run_args)
            else:
                popen_args.extend(self.config.run_args)
        self.osgi_runner = subprocess.Popen(popen_args)
        return self

    def __exit__(
        self,
        __exc_type: t.Optional[t.Type[BaseException]],
        __exc_value: t.Optional[BaseException],
        __traceback: t.Optional[TracebackType],
    ) -> t.Optional[bool]:
        osgi_runner_ = self.osgi_runner

        if osgi_runner_ is None:
            log.warn("Called __exit__ on inactive OSGIRunner.")
            return None

        osgi_runner = osgi_runner_

        log.info("Stopping osgi-runner with PID %d upon tear down." % osgi_runner.pid)
        osgi_runner.terminate()
        log.info("Waiting for osgi-runner to stop upon tear down.")
        rc = retry_operation_simple(
            "stop osgi-runner", lambda: osgi_runner.wait(1), ntries=20, interval=0.1
        )
        log.info("osgi-runner stopped with return code [%s] upon tear down." % rc)
        # log.info("osgiRunner.kill() ...")
        # cls.osgiRunner.kill()
        # log.info("osgiRunner.communicate() ...")
        # cls.osgiRunner.communicate()
        # log.info("None-ing osgiRunner")
        self.osgi_runner = None
        log.info("Done tearing down osgi-runner")
        return None


@dataclass
class PsycopgConnectionConfig:
    dsn: str
    user: str
    password: str
    ntries: int = 30
    readonly: bool = False
    autocommit: bool = True


class PsycopgConnection(AbstractContextManager[psycopg2.extensions.connection]):
    config: PsycopgConnectionConfig
    connection: t.Optional[psycopg2.extensions.connection]

    def __init__(self, config: PsycopgConnectionConfig):
        self.config = config
        self.connection = None

    def __enter__(self) -> psycopg2.extensions.connection:
        self.connection = retry_operation_simple(
            "db connection",
            lambda: psycopg2.connect(
                self.config.dsn, user=self.config.user, password=self.config.password
            ),
            ntries=self.config.ntries,
        )
        self.connection.set_session(
            readonly=self.config.readonly, autocommit=self.config.autocommit
        )
        return self.connection

    def __exit__(
        self,
        __exc_type: t.Optional[t.Type[BaseException]],
        __exc_value: t.Optional[BaseException],
        __traceback: t.Optional[TracebackType],
    ) -> t.Optional[bool]:
        if self.connection is None:
            log.warn("Called __exit__ on inactive PsycopgConnection.")
            return None

        self.connection.close()
        self.connection = None
        return None

# The client-side mTLS certificate is represented using request headers,
# because the osgi login classes assume that we are behind a tls terminating reverse proxy
# that puts the information contained in the client side certificate into request headers.
CertificateHeaders: t.TypeAlias = dict[str,str]

def add_http_errors(s: requests.Session) -> requests.Session:
    """
    Automatically call requests.Response.raise_for_status for every request using this session.
    """
    s.hooks.setdefault("response", []).append(lambda r, **kwargs: r.raise_for_status())
    return s

#
def add_login_session(s: requests.Session, login_info: t.Optional[LoginInfo | CertificateHeaders] = None) -> requests.Session:
    """
    Create a requests.Session that is logged in using the provided LoginInfo or using fake mTLS headers.
    """
    if login_info is None:
        pass
    elif isinstance(login_info, LoginInfo):
        s.cookies.update(login_info.cookies) # type: ignore[no-untyped-call]
    else:
        s.headers.update(login_info)

    return s

# We need to define this outselves because the type of the assertRaises context manager isn't exported.
class _AssertRaisesContext(t.Protocol, AbstractContextManager["_AssertRaisesContext"]):
    exception: t.Union[XmlRpcTargetError, requests.HTTPError]

class AssertNeedsLogin(AbstractContextManager["AssertNeedsLogin"]):
    """
    An assertRaises derivative that asserts that a request that was executed in this context, is asking for a login from the user.

    If the needs_login_url parameter is False: We instead assert that the response is asking for a non-interactive login/authentication.
    """

    _testcase: unittest.TestCase
    _assert_raises: _AssertRaisesContext
    needs_login_url: bool
    def __init__(self, testcase: unittest.TestCase, needs_login_url: bool = True):
        self._testcase = testcase
        self._assert_raises = testcase.assertRaises((XmlRpcTargetError, requests.HTTPError)) # important
        self.needs_login_url = needs_login_url
    def __enter__(self) -> "AssertNeedsLogin":
        self._assert_raises.__enter__()
        return self

    def __exit__(
        self,
        exc_type: t.Optional[t.Type[BaseException]],
        exc_value: t.Optional[BaseException],
        traceback: t.Optional[TracebackType],
    ) -> t.Optional[bool]:
        # start of important section
        ret = self._assert_raises.__exit__(exc_type, exc_value, traceback)
        if ret:
            has_login_url: bool
            if isinstance(self._assert_raises.exception, XmlRpcTargetError):
                self._testcase.assertEqual(JSON_RPC_LOGIN_REQUIRED,self._assert_raises.exception.code)
                data_value = self._assert_raises.exception.data
                has_login_url = isinstance(data_value, dict) and isinstance(data_value.get("loginUrl"), str) and bool(data_value["loginUrl"])
            else:
                self._testcase.assertEqual(401,self._assert_raises.exception.response.status_code)
                has_login_url = "X-HTTP-util-login-url" in self._assert_raises.exception.response.headers

            self._testcase.assertTrue(self.needs_login_url == has_login_url, "Expected a login url" if self.needs_login_url else "No login url expected")
        return ret
        # end of important section

class AssertForbidden(AbstractContextManager["AssertForbidden"]):
    """
    Asserts that a request.HTTPError with status 403 was raised or the corresponding xmlrpc exception.
    """
    _testcase: unittest.TestCase
    _assert_raises: _AssertRaisesContext

    def __init__(self, testcase: unittest.TestCase):
        self._testcase = testcase
        self._assert_raises = testcase.assertRaises((XmlRpcTargetError, requests.HTTPError)) # important

    def __enter__(self) -> "AssertForbidden":
        self._assert_raises.__enter__()
        return self

    def __exit__(
        self,
        exc_type: t.Optional[t.Type[BaseException]],
        exc_value: t.Optional[BaseException],
        traceback: t.Optional[TracebackType],
    ) -> t.Optional[bool]:
        # start of important section
        ret = self._assert_raises.__exit__(exc_type, exc_value, traceback)
        if ret:
            if isinstance(self._assert_raises.exception, XmlRpcTargetError):
                self._testcase.assertEqual(JSON_RPC_SERVER_ERROR_0 - 1, self._assert_raises.exception.code)
            else:
                self._testcase.assertEqual(403, self._assert_raises.exception.response.status_code)
        return ret
        # end of important section

def assertAllNeedLogin(testcase: unittest.TestCase, bean: Bean) -> None:
    '''
    Checks, whether all bean calls of the given service bean need a login
    by invoking ~assertNeedsLogin.
    '''
    log.info("Checking all methods of service [%s] for a required login..."%(bean,))

    for fn in bean.methods:
        f = getattr(bean,fn)

        with AssertNeedsLogin(testcase):
            f(*f.info.getDefaultParameters())

    log.info("[OK] All methods of service [%s] require a login."%(bean,))
