'''
  OSGi python tools of the clazzes.org project
  http://www.clazzes.org

  Created: 07.04.2020

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

      http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
'''

import re
import logging

from paramiko.client import SSHClient, MissingHostKeyPolicy
from paramiko.util import ClosingContextManager
from paramiko.pkey import PKey
from paramiko.rsakey import RSAKey
import subprocess
import io
from dataclasses import dataclass
import typing as t
from enum import Enum, auto

log=logging.getLogger(__name__)

class OsgiShellError(Exception):
    pass

class IdNotFoundException(Exception):
    pass

BUNDLE_NAME_REGEX = re.compile(r"\s*\d{1,3}\|\w+\s*\|\w*\s*\|\s*\d{1,3}\|([^\r]*)")
GOGO_FIND_TABLE_AMOUNT_REGEX = re.compile(r'^(\d+)\sresults?(\r\n)?')

@t.overload
def findBundleId(bundleName:str,bundleList:t.Sequence[str], force:t.Literal[True]) -> int: ...
@t.overload
def findBundleId(bundleName:str,bundleList:t.Sequence[str], force:bool = False) -> t.Optional[int]: ...

def findBundleId(bundleName:str,bundleList:t.Sequence[str], force:bool = False) -> t.Optional[int]:
    """
    Searches the list of bundles returned by the gogo-shell for the id of the bundle with the given name.

    @param bundleName: The name of the bundle to be searched for
    @param bundleList: The list of currently registered bundles. Each line is one entry in the list and has the format:
    <BUNDLE-ID>|[Active/Resolved]     |(CREATED)     |  <BUNDLE-PRIORITY>|<BUNDLE-NAME>
    ...where <BUNDLE-ID>, <BUNDLE-PRIORITY> and <BUNDLE-NAME> are substituted by the corresponding value, (CREATED) means
    that the string 'CREATED' is optionally present and [Active/Resolved] means that one of those strings is present.
    :param force: If True, throw an IdNotFoundException istead of returning None, when
                  the given bundle name could not be found.
    @return: The id of the bundle, or None if it wasn't found
    """
    #Typical line returned by the 'ls' command:
    #[BUNDLE-ID]|Active     |CREATED     |  [BUNDLE-PRIORITY]|[BUNDLE-NAME]

    bundleRegexRawString = r'^\s*(\d+)\|\w+\s*\|\w*\s*\|\s*\d+\|' + re.escape(bundleName)
    bundleRegex = re.compile(bundleRegexRawString)

    bundleId = None
    for line in bundleList:
        r = bundleRegex.match(line)
        if r:
            bundleId = int(r.group(1))

    if force and bundleId is None:
        raise IdNotFoundException("ID for bundle [%s] could not be found."%(bundleName,))

    return bundleId

class IgnorePolicy(MissingHostKeyPolicy):
    def missing_host_key(self, client: SSHClient, hostname: str, key: PKey) -> None:
        pass

class KnownHostsStrategy(Enum):
    # Load the system known hosts file.
    SYSTEM = auto()
    # Connect irregardless of any known hosts.
    IGNORE = auto()

@dataclass
class OSGIClientConfig:
    # If None use ssh agent.
    ssh_port: int = 8101
    hostname: str = "localhost"
    ssh_key: t.Optional[PKey] = None
    ssh_key_file: t.Optional[str] = None
    # If None use system known_hosts.
    known_hosts: t.Union[str, KnownHostsStrategy] = KnownHostsStrategy.SYSTEM
    datasource_engine: t.Optional[str] = "MYSQL"
    datasource_name: t.Optional[str] = None
    datasource_schema: t.Optional[str] = None
    datasource_catalog:t.Optional[str] = None


class OSGiClient(ClosingContextManager):
    config: OSGIClientConfig
    client: SSHClient

    @t.overload
    def __init__(self,
                 ssh_port:int=8101,
                 ssh_key:t.Optional[PKey]=None,
                 known_hosts:t.Optional[str]=None,
                 datasourceEngine:t.Optional[str]="MYSQL",
                 datasourceName:t.Optional[str]=None,
                 datasourceSchema:t.Optional[str]=None,
                 datasourceCatalog:t.Optional[str]=None): ...

    @t.overload
    def __init__(self,config:OSGIClientConfig): ...

    def __init__(self, # type: ignore[misc]
                 ssh_port:t.Union[int,OSGIClientConfig]=8101,
                 ssh_key:t.Optional[PKey]=None,
                 known_hosts:t.Optional[str]=None,
                 datasourceEngine:t.Optional[str]="MYSQL",
                 datasourceName:t.Optional[str]=None,
                 datasourceSchema:t.Optional[str]=None,
                 datasourceCatalog:t.Optional[str]=None):
        if isinstance(ssh_port, OSGIClientConfig):
            self.config = ssh_port
        else:
            self.config = OSGIClientConfig(
                ssh_key=ssh_key,
                ssh_port=ssh_port,
                known_hosts=KnownHostsStrategy.SYSTEM if known_hosts is None else known_hosts,
                datasource_engine=datasourceEngine,
                datasource_name=datasourceName,
                datasource_schema=datasourceSchema,
                datasource_catalog=datasourceCatalog)

        self.client = SSHClient()

    @property
    def ssh_port(self)->int:
        return self.config.ssh_port
    @property
    def known_hosts(self)->t.Optional[t.Union[str, KnownHostsStrategy]]:
        return self.config.known_hosts
    @property
    def datasourceEngine(self)->t.Optional[str]:
        return self.config.datasource_engine
    @property
    def datasourceName(self)->t.Optional[str]:
        return self.config.datasource_name
    @property
    def datasourceSchema(self)->t.Optional[str]:
        return self.config.datasource_schema
    @property
    def datasourceCatalog(self)->t.Optional[str]:
        return self.config.datasource_catalog

    def __enter__(self)->"OSGiClient":
        self.connect()
        return self

    def connect(self)->None:

        if self.known_hosts is None or self.known_hosts == KnownHostsStrategy.SYSTEM:
            self.client.load_system_host_keys()
        elif self.known_hosts == KnownHostsStrategy.IGNORE:
            self.client.set_missing_host_key_policy(IgnorePolicy());
        else:
            assert not isinstance(self.known_hosts, KnownHostsStrategy)
            self.client.load_host_keys(self.known_hosts)

        if self.config.ssh_key is not None:
            ssh_key = self.config.ssh_key
        elif self.config.ssh_key_file is not None:
            ssh_key = RSAKey.from_private_key_file(self.config.ssh_key_file)
        else:
            ssh_key = None

        self.client.connect(self.config.hostname,port=self.ssh_port, pkey=ssh_key,allow_agent=ssh_key is None)

    def close(self)->None:
        self.client.close()

    def runCommand(self,cmd: str)->t.Tuple[str,...]:

        _, stdout,stderr = self.client.exec_command(cmd)

        # read stdout first, because it might contain a lot of logging information and
        # reading stderr would block in this situation.
        lines = stdout.readlines()
        errors = stderr.readlines()

        if errors:
            raise OsgiShellError("OSGi command [%s] failed:"%cmd,*errors)

        ret = tuple(line.rstrip('\r\n') for line in lines)

        # filter out one single empty line as response.
        if len(ret) == 1 and ret[0] == '':
            return ()

        return ret

    def getBundleList(self) -> t.List[str]:

        bundleList = self.runCommand('ls -l')

        bundleNameList = []

        for line in bundleList:
            m = BUNDLE_NAME_REGEX.match(line)

            if m:
                bundleNameList.append(m.group(1))

        return bundleNameList

    def countDBTables(self) -> int:
        """
        Via osgi-shell get the count of the tables from datasource
        possibly filtered by either catalog or schema (if configured)
        """
        if self.datasourceName is None:
            raise ValueError("OSGi datasource name is not configured.")
        log.debug(f"Counting tables of database {self.datasourceName}")
        if self.datasourceEngine == "MYSQL":
            log.debug("Assuming MYSQL database...")
            if self.datasourceSchema is None:
                if self.datasourceCatalog is None:
                    cmd = "showtables %s"%self.datasourceName
                else:
                    cmd = "showtables %s -C %s" % (self.datasourceName, self.datasourceCatalog)
            else:
                if self.datasourceCatalog is None:
                    cmd = "showtables %s -s %s"%(self.datasourceName, self.datasourceSchema)
                else:
                    cmd = "showtables %s -C %s -s %s"%(self.datasourceName, self.datasourceCatalog, self.datasourceSchema)
            #The results count should be the last line of the output
            lines = self.runCommand(cmd)
            log.debug(f"Output from count tables statement: {lines}")

            r = GOGO_FIND_TABLE_AMOUNT_REGEX.match(lines[-1])
            #If regex didn't match, unexpected amount of tables
            if r:
                return int(r.group(1))
            else:
                raise ValueError("OSGi command [%s] returned unexpected output: %s" % (cmd,lines[-1]))

        elif self.datasourceEngine == "PSQL":
            log.debug("Assuming PSQL database...")
            if not self.datasourceCatalog:
                log.error("Datasource catalog is required to count tables in a PSQL database.")
                raise ValueError("Datasource catalog is required to count tables in a PSQL database.")
            cmd = f"SELECT COUNT(*) FROM information_schema.tables WHERE table_catalog='{self.datasourceCatalog}' "
            if self.datasourceSchema:
                cmd += f"AND table_schema='{self.datasourceSchema}';"
            else:
                # Assuming everyting is in 'public',
                cmd += "AND table_schema='public';"
            cmd = f"query {self.datasourceName} \"{cmd}\""
            log.debug(f"Running cmd '{cmd}'")
            lines = self.runCommand(cmd)
            log.debug(f"Output from count tables statement: {lines}")
            # 'X results' is the default output for queries in PSQL
            if len(lines) < 2 or not re.match(r"\d+ results?", lines[-1]):
                raise ValueError("OSGi command [%s] returned unexpected output: %s" % (cmd,lines))
            # This should be the actual result of the 'count' column
            r = re.match(r"\|?(\d+)", lines[-2])
            if r:
                return int(r.group(1))
            else:
                raise ValueError("OSGi command [%s] returned unexpected output: %s" % (cmd,lines))
        else:
            log.error("Cannot count tables without `datasourceEngine` parameter.")
            raise ValueError("Cannot count tables without `datasourceEngine` parameter.")


    def getLatestSchemaVersion(self) -> t.Optional[str]:
        if self.datasourceName is None:
            raise ValueError("OSGi datasource name is not configured.")

        cmd = f"jdbc:query --raw {self.datasourceName} \"SELECT VERSION FROM SCHEMA_HISTORY ORDER BY SERIALNR DESC OFFSET 0 ROWS FETCH NEXT 1 ROWS ONLY\"";

        lines = self.runCommand(cmd)

        if len(lines) <= 0:
            return None

        return lines[0]


    def findBundleId(self,bundleName:str, force:bool = False) -> t.Optional[int]:
        bundleList = self.runCommand('ls')
        ret = findBundleId(bundleName,bundleList,force)
        return ret

    @t.overload
    def findBundleIds(self,bundleNames:t.Sequence[str], force:t.Literal[True]) -> t.Tuple[t.Tuple[int,str],...]: ...
    @t.overload
    def findBundleIds(self,bundleNames:t.Sequence[str], force:bool = False) -> t.Tuple[t.Tuple[t.Optional[int],str],...]: ...

    def findBundleIds(self,bundleNames:t.Sequence[str], force:bool = False) -> t.Tuple[t.Tuple[t.Optional[int],str],...]:
        bundleList = self.runCommand('ls')
        return tuple((findBundleId(bundleName,bundleList,force),bundleName) for bundleName in bundleNames)

    def stopBundle(self,bundleId:int,bundleName:str)->None:
        log.info("Stopping bundle[%d/%s]..."%(bundleId,bundleName))
        lines=self.runCommand("stop %d"%bundleId)
        for line in lines:
            log.info(line)

    def startBundle(self,bundleId:int,bundleName:str)->None:
        log.info("Starting bundle[%d/%s]..."%(bundleId,bundleName))
        lines=self.runCommand("start %d"%bundleId)
        for line in lines:
            log.info(line)

    def dropTable(self,dbTable:str)->None:
        log.info("Dropping table [%s]..."%dbTable)
        lines=self.runCommand("query %s drop table %s"%(self.datasourceName,dbTable))
        for line in lines:
            log.info(line)

    def dropFunction(self,dbFunction:str)->None:
        log.info("Dropping function [%s]..."%dbFunction)
        lines=self.runCommand("query %s drop function %s"%(self.datasourceName,dbFunction))
        for line in lines:
            log.info(line)


    def restartDB(self,bundleNames:t.Sequence[str],dbTables:t.Optional[t.Sequence[str]]=None,script:t.Optional[t.Sequence[str]]=None,input:t.Optional[bytes]=None,inputFile:t.Optional[str]=None,dbFunctions:t.Optional[t.Sequence[str]]=None)->None:
        '''
           Stop the given bundles in reverse order, drop the given DB tables and/or
           run the given script and then restart the DB.

           :param bundleNames: A list of bundles to start in order. The sort order is reversed.
           :param dbTables: A list of DB table to drop.
           :param script: A list of parameters for a script used to reset the DB after stopping the bundles.
           :param input: An input byte array for the script to execute.
           :param inputFile: A file to feed as input to the script, if input is None.
           :param dbFunctions: A list of DB functions to drop.
        '''
        if self.datasourceName is None:
            raise ValueError("OSGi datasource name is not configured.")

        bundles = self.findBundleIds(bundleNames,True)
        self.stop_bundles(reversed(bundles))

        if dbTables:
            for dbTable in dbTables:
                self.dropTable(dbTable)

        if dbFunctions:
            for dbFunction in dbFunctions:
                self.dropFunction(dbFunction)

        if script:

            if input is None and inputFile is not None:
                with io.open(inputFile,"rb") as fh:
                    input = fh.read()

            log.info("Running script %s."%(script,))
            subprocess.run(script,check=True,input=input)

        self.start_bundles(reversed(bundles))

    def stop_bundles(self, bundles: t.Iterable[t.Tuple[int,str]])->None:
        for bundle in bundles:
            self.stopBundle(*bundle)

    def start_bundles(self, bundles: t.Iterable[t.Tuple[int,str]])->None:
        for bundle in bundles:
            self.startBundle(*bundle)
