import os.path
import io
import json
import typing as t
import requests
import urllib.parse
from datetime import datetime, timezone, timedelta
from dataclasses import dataclass, field
import traceback
import logging
from contextlib import AbstractContextManager
from types import TracebackType
import yaml
from typing import List
import pathlib
import sys

log = logging.getLogger(__name__)


STATUS_Passed = "Pass"
STATUS_Failed = "Fail"
STATUS_Blocked = "Blocked"
STATUS_Untested = "Not Executed"
STATUS_Ongoing = "In Progress"


ZEPHYR_API_URL = "https://api.zephyrscale.smartbear.com/v2"

@dataclass
class ZephyrStepResult:
    """
        A zephyr step result.
    """

    statusName: str
    actualEndDate: t.Optional[str] = None
    actualResult: t.Optional[str] = None

@dataclass
class ZephyrResult:
    """
       A zephyr result to POST as testrun result.
    """
    projectKey: str
    testCaseKey: str
    testCycleKey: str
    statusName: str
    testScriptResults: t.Optional[t.Sequence[ZephyrStepResult]] = None,
    actualEndDate: t.Optional[str] = None
    executionTime: t.Optional[int] = None
    comment: t.Optional[str] = None

class ZephyrEncoder(json.JSONEncoder):
    def default(self, d:object) -> object:
        if isinstance(d, (ZephyrResult, ZephyrStepResult, ZephyrAccess)):
            return d.__dict__
        else:
            return super().default(d)
    

@dataclass
class ZephyrAccess:
    """
    ~/.config/clazzes/zephyr/tokens.json

   :ivar serverId: The ID of the server used on command line arguments, mostly
                    a string like 'clazzes' or 'iteg'.
   :ivar apiKey: The API key for the given user for authenticating against zephyr.
    """
    
    serverId: str
    apiKey: str

    def performRequest(self, *urlParts:object, queryParams:t.Dict[str, str] = {}, payload:object=None) -> requests.Response:
        """
           Perform an API request by concatenating all positional arguments with
           a slash, turning them into path parameters.
           By default, send a GET request. If a payload exists, instead send a POST request with 
           the payload.
           As of now, does not implement URL encoded parameters
           

            :param urlParts: Parts of the REST API call which should be appended to the base URL and concatenated by `/`. 
            :param queryParams: Appended at the end.
            :param payload: t.Optional payload for POST requests, which is JSON-encoded. If active, set content-type 
                                to application/json
            :return: A response object with a postive HTTP status code.
            :rtype: requests.Response
        """

        url = ZEPHYR_API_URL
        requestHeaders = {
            "accept": "application/json",
            "Authorization": f"Bearer {self.apiKey}"
        }

        for param in urlParts:
            url += "/"
            if isinstance(param,str):
                url += urllib.parse.quote(param,safe=())
            else:
                url += urllib.parse.quote(str(param),safe=())

        if queryParams:
            url += "?"
            query_string = "&".join([f"{urllib.parse.quote(str(param))}={urllib.parse.quote(str(value))}" for param, value in queryParams.items()])
            url += query_string

        log.info(f"Performing request to [{url}]")

        if payload is None:
            resp = requests.get(url, headers=requestHeaders)
        else:
            data = json.dumps(payload, cls=ZephyrEncoder)

            if log.getEffectiveLevel() <= logging.DEBUG:
                log.debug(f"POSTing data [{data}]")
                
            #Add content type for POSTing json data
            requestHeaders["content-type"] = "application/json"

            resp = requests.post(url,
                                 data=data,
                                 headers=requestHeaders
                                 )

        log.info(f"Request to [{url}] returned [{resp.status_code}]")

        if 400 <= resp.status_code < 500 or 500 <= resp.status_code < 600:
            log.error(f"Body of request to [{url}] was:\n{resp.text}")

        resp.raise_for_status()
        return resp

    def addResultForCase(self, result: ZephyrResult) -> requests.Response:
        """
        POST a zephyr result to /api/v2/add_result_for_case/{run_id}/{case_id}

       :param result: Data object containing all result data, including the identifiers
                       of both the test case and test cycle to add the result to.
       :type result: ZephyrResult
       :return: A response object with a postive HTTP status code.
       :rtype: requests.Response
        """
        return self.performRequest("testexecutions", payload=result)

@dataclass
class ZephyrTestCase(AbstractContextManager):
    testCaseKey: str
    stepCount: int
    projectKey: t.Optional[str] = None
    testCycleKey: t.Optional[str] = None
    name: t.Optional[str] = None
    access: t.Optional[ZephyrAccess] = None
    reentrant: t.Optional[bool] = False
    start: datetime = datetime.now().replace(tzinfo=timezone.utc)
    end: t.Optional[datetime] = None
    elapsed: t.Optional[timedelta] = None
    testScriptResults: list[ZephyrStepResult] = field(default_factory=list)
    zephyrApiError: t.Optional[bool] = False

    def commit(self, result: ZephyrResult)->None:
        # set the committed flag in order to not propagate Zephyr errors
        self.end = datetime.now().replace(tzinfo=timezone.utc)
        self.elapsed = self.end - self.start
        #Add timing information (execution time in milliseconds, end date in formatted string)
        result.executionTime = self.elapsed / timedelta(milliseconds = 1)
        result.actualEndDate = self.end.isoformat()

        # set the status on the server.
        if self.access is not None and self.testCycleKey is not None:
            #log.info(f"Propagating zephyr result [{result}] for case [{self.testCaseKey}]")
            try:
                self.access.addResultForCase(result)
                self.zephyrApiError = False
            except:
                self.zephyrApiError = True
                raise
        else:
            log.warning(f"Not propagating zephyr result [{result}] for case [{self.testCaseKey}]")

    def addStepResult(self, statusName:int, actualResult: t.Optional[str] = None, actualEndDate: t.Optional[str] = None)->None:
        result = ZephyrStepResult(statusName, actualEndDate, actualResult)
        if (not self.testScriptResults):
            self.testScriptResults =  [result] 
        else:
            self.testScriptResults.append(result)

    def passed(self,comment:str)->None:
        result = ZephyrResult(self.projectKey,
                              self.testCaseKey,
                              self.testCycleKey,
                              STATUS_Passed,
                              testScriptResults=self.testScriptResults,
                              comment=comment)
        # Execution time and actual end results are added in the commit.
        self.commit(result)

    def failed(self,comment:str)->None:
        #Zephyr API requires the number of steps to match always, so fill up the difference
        blockedResult = ZephyrStepResult(STATUS_Blocked, datetime.now().replace(tzinfo=timezone.utc).isoformat())
        remainingResults = [blockedResult for _ in range(len(self.testScriptResults), self.stepCount)]
        self.testScriptResults.extend(remainingResults)
        result = ZephyrResult(self.projectKey,
                              self.testCaseKey,
                              self.testCycleKey,
                              STATUS_Failed,
                              testScriptResults=self.testScriptResults,
                              comment=comment)
        self.commit(result)

    def __enter__(self)->"ZephyrTestCase":
        return self

    def __exit__(
        self,
        exc_type: t.Optional[t.Type[BaseException]],
        exc_val: t.Optional[BaseException],
        exc_tb: t.Optional[TracebackType],
    ) -> bool:

        # If zephyrApiError is set to True, then the exception was caught when trying to connect to Zephyr.
        # Such exceptions are propagated "as is" by returning False (causes exceptions to be re-thrown)
        if self.zephyrApiError == True:
            return False

        if self.end is None:

            if exc_type is None and exc_val is None and exc_tb is None:
               raise ValueError(f"Test case [{self.testCaseKey}] has not been committed, please fix your code and add tc.passed() or tc.failed().")

            comment = "".join(traceback.format_exception(exc_type, exc_val,
                                          exc_tb))
            try:
                self.failed(comment)
            except:
                log.exception("Error propagating exception to zephyr")
                return False

        return self.reentrant

    def step(self)->"ZephyrTestStep":
        return ZephyrTestStep(self)


@dataclass
class ZephyrTestStep(AbstractContextManager):
    testCase: ZephyrTestCase
    statusName: str = STATUS_Untested
    committed: bool = False
    start: datetime = datetime.now().replace(tzinfo=timezone.utc)
    end: t.Optional[datetime] = None

    def commit(self, statusName: str, actualResult: t.Optional[str])->None:
        self.committed = True
        self.end = datetime.now().replace(tzinfo=timezone.utc)
        self.testCase.addStepResult(statusName, actualResult, self.end.isoformat())
        
    def passed(self, actualResult: t.Optional[str] = None):
        self.commit(STATUS_Passed, actualResult)
    
    def failed(self, actualResult: str):
        self.commit(STATUS_Failed, actualResult)
        
    def __enter__(self)->"ZephyrTestStep":
        return self
    
    def __exit__(
        self,
        exc_type: t.Optional[t.Type[BaseException]],
        exc_val: t.Optional[BaseException],
        exc_tb: t.Optional[TracebackType],
    ) -> t.Literal[False]:
        if not self.committed:
            if exc_type is None and exc_val is None and exc_tb is None:
                # In this case, the step was finished without exceptions, but without using passed() first -> remind user
                raise ValueError(f"Step of Test case [{self.testCase.testCaseKey}] has not been committed, please fix your code and add ts.passed() or ts.failed().")
            else:
                # In this case, an Exception(likely AssertionError) was thrown within the step context, so we commit a failure
                self.actualResult = f"{exc_val}"

                self.failed(self.actualResult)
                return False

        #If step has already been committed without issue, exit cleanly
        return False


def readToken(filename: str)->t.List[ZephyrAccess]:
    """
    Obtain Zephyr API access via the token.
    """

    log.info(f"Loading list of API tokens from file [{filename}].")

    with open(filename,"r",encoding='utf-8') as fh:
        loaded = json.load(fh,object_hook=lambda x: ZephyrAccess(**x))
        if isinstance(loaded, list):
            return loaded
        else:
            raise TypeError("Json isn't a list.")


def readUserTokens()->t.List[ZephyrAccess]:
    """
    Look in user config folder for zephyr API key.
    """

    fn = os.path.expanduser('~/.config/clazzes/zephyr/servers.json')

    if os.path.exists(fn):
        return readToken(fn)
    else:
        log.warning(f"File [{fn}] does not exists, add a content like follows to it:")
        log.warning("[")
        log.warning('  { "serverId": "clazzes",')
        log.warning('    "apiKey": "<api-key>"')
        log.warning('  }')
        log.warning("]")
        return []

def getAccess(servers:t.Sequence[ZephyrAccess],serverId:str) -> t.Optional[ZephyrAccess]:
    for server in servers:
        if server.serverId == serverId:
            log.info(f"Token for server [{serverId}] obtained successfully.")
            return server

    raise ValueError(f"Cannot find server with ID [{serverId}] in list of servers.")

def getUserToken(serverId:str) -> t.Optional[ZephyrAccess]:
    return getAccess(readUserTokens(),serverId)


@dataclass
class ZephyrTestCycleDefinition:
    cycleName: str #This is NOT the ID/Key of the test cycle, but its name
    caseKeys: List[str]
            

@dataclass
class ZephyrTestSuiteDefinition:
    projectKey: str
    cycles: List[ZephyrTestCycleDefinition]
    
    @classmethod
    def from_yaml(cls, yaml_content):
        data = yaml.safe_load(yaml_content)
        cycles = [ZephyrTestCycleDefinition(cycleName=name, caseKeys=cases) for name, cases in data["cycles"].items()]
        return cls(data['projectId'], cycles)
        
    
def setupNewTestPlan(access: ZephyrAccess, definitionPath: str, version : str):

    with open(definitionPath) as f:
        yaml_content = f.read()
        
    testSuiteDef = ZephyrTestSuiteDefinition.from_yaml(yaml_content)
    
    # Create test plan
    createPlanData = {
        "projectKey": testSuiteDef.projectKey,
        "name": f"Release {version}",
        "labels": ["Regression"]
        #"statusName": STATUS_Ongoing
    }
    resp = access.performRequest("testplans", payload=createPlanData)
    testPlanId = resp.json()["id"]
        
    # Create test cycles
    # First, create new folder
    
    createFolderData = {
        "name" : version,
        "projectKey" : testSuiteDef.projectKey,
        "folderType" : "TEST_CYCLE"
    }
    resp = access.performRequest("folders", payload=createFolderData)
    folderId = resp.json()["id"]
        
    # Place all the new test cycles in the new folder
    
    for cycle in testSuiteDef.cycles:
        createCycleData = {
            "projectKey" : testSuiteDef.projectKey,
            "name" : cycle.cycleName,
            "statusName" : STATUS_Untested,
            "folderId" : folderId,
        }
        resp = access.performRequest("testcycles", payload=createCycleData)
        testCycleKey = resp.json()["key"]
        # Link the test cycle to the test plan
        log.debug("Created test cycle with key %s", testCycleKey)
        
        createLinkData = {
            "testCycleIdOrKey" : testCycleKey
        }
        access.performRequest("testplans", testPlanId, "links", "testcycles", 
                              payload=createLinkData)
        
        # Finally, add the test cases to the test cycles
        # which is done by making "Not Executed" test executions, which are effectively placeholders
        for caseKey in cycle.caseKeys:
            addPlaceholderExecData = {
                "projectKey": testSuiteDef.projectKey,
                "testCaseKey": f"{testSuiteDef.projectKey}-{caseKey}",
                "testCycleKey": testCycleKey,
                "statusName": STATUS_Untested
                # Note that the "testScriptResults" property is missing
            }
            access.performRequest("testexecutions", payload=addPlaceholderExecData)
        
    log.info(f"Finished creating test plan for version {version} with id {testPlanId}")
    
@dataclass
class ZephyrConfig:
    """
    A configuration object storing everything necessary for a test run execution
    
    :ivar zephyrAccess: Stores login information to the Zephyr API
    :ivar applicationVersion: Version of the release that is tested. Used when creating test plans.
    :ivar testDef: Stores references to the Jira Zephyr infrastructure objects (project id, test cycle names and cases). 
                   Is parsed from a yaml file that can vary between projects.
    :ivar testCycleFolder: 
    :ivar testCycleIdsByName: Stores references to test cycle ids (which are non-static, unlike the test cycle names 
    defined in the testDef and differ between test executions)
    """
    zephyrAccess: t.Optional[ZephyrAccess] = None
    applicationVersion: t.Optional[str] = None
    testDef : t.Optional[ZephyrTestSuiteDefinition] = None
    testCycleFolder: t.Optional[str] = None
    testCycleKeysByName : dict[str, str] = field(default_factory=dict)
    
    def case(self, cycleName: str, caseKey: str, stepCount: int, reentrant: bool = False, name: str = None) -> ZephyrTestCase:
        
        testCycleKey = None
        if self.testDef:
            testCycleKey = self.testCycleKeysByName[cycleName]
        else:
            log.info("testDef not set. Will not propagate to Zephyr instance.")
            
        return ZephyrTestCase(f"{self.testDef.projectKey}-{caseKey}" if self.testDef else caseKey,
                              stepCount,
                              self.testDef.projectKey if self.testDef else None, 
                              testCycleKey, 
                              name,
                              self.zephyrAccess, 
                              reentrant, 
                              start=datetime.now().replace(tzinfo=timezone.utc))
        
    def loadTestDefFromYaml(self, yamlContent: str):
        self.testDef = ZephyrTestSuiteDefinition.from_yaml(yamlContent)
        
    def fetchCycleKeysForNames(self):
        if not self.testDef:
            raise ValueError("`testDef` property not set, cannot obtain cycles from server")
        
        folderName = self.testCycleFolder if self.testCycleFolder else self.applicationVersion

        if not folderName:
            raise ValueError("Cannot fetch test cycles, neither folder name nor applicationVersion are specified.")
        
        log.info("Attempting to fetch test cycles in folder %s", folderName)
        #First, get the folder of the release
        resp = self.zephyrAccess.performRequest("folders", queryParams={"folderType" : "TEST_CYCLE"})
        folders = resp.json()["values"]
        releaseFolder = next((f for f in folders if folderName == f["name"]), None)
        
        if not releaseFolder:
            raise ValueError(f"Folder [${folderName}] not found on Zephyr Test Cycle Folder structure in the root folder.")
        
        #Fetch test cycles in the folder
        testCycleParams = {
            "projectKey" : self.testDef.projectKey,
            "folderId" : releaseFolder["id"]
        }
        resp = self.zephyrAccess.performRequest("testcycles", queryParams=testCycleParams)
        cyclesByName = {c["name"] : c["key"]  for c in resp.json()["values"]}
        log.info("Obtained test cycle mapping: %s", cyclesByName)
        if set(cyclesByName.keys()) != {c.cycleName for c in self.testDef.cycles}:
            raise ValueError("Test cycles names are different to those configured in the yaml...")
        
        #Adopt the new testcycle map
        self.testCycleKeysByName = cyclesByName

### Tests ###

def testAddingResults(conf: ZephyrConfig):
    try:
        with (conf.case("DDSAGILE-T105") as case):
            with ZephyrTestStep(case) as ts:
                ts.passed("Worked fine")
            with ZephyrTestStep(case) as ts:
                ts.passed("Worked fine")
            with ZephyrTestStep(case) as ts:
                print("Forgot to commit... Awaiting exception")
    except Exception as e:
        print(f"Caught exception, as expected: [{e}]")
    
    try:
        with (conf.case("DDSAGILE-T99") as case):
            with ZephyrTestStep(case) as ts:
                ts.passed("Worked fine")
            with ZephyrTestStep(case) as ts:
                ts.passed("Worked fine")
            with ZephyrTestStep(case) as ts:
                ts.passed("Worked fine")
            with ZephyrTestStep(case) as ts:
                ts.failed("Oops, had a problem")
            print("Not commiting the test case, should catch exception")
    except Exception as e:
        print(f"Caught exception as expected: [{e}]")
        
    with (conf.case("DDSAGILE-T99") as case):
        with ZephyrTestStep(case) as ts:
            ts.passed("Worked fine")
        with ZephyrTestStep(case) as ts:
            ts.passed("Worked fine")
        with ZephyrTestStep(case) as ts:
            ts.passed("Yay!")
        case.passed("Everything fine!")   
        
def testAddingNewTestPlan(conf:ZephyrConfig, version: str):
    scriptDir = pathlib.Path(__file__).parent.resolve()
    setupNewTestPlan(conf.zephyrAccess, scriptDir / "zephyr_cfg.yml", version)

if __name__ == "__main__":
    log.setLevel("DEBUG")
    log.addHandler(logging.StreamHandler(sys.stdout))
    access = getUserToken("ITEG-Atlassian")
    conf = ZephyrConfig(access, "4.1.0")
    with open("/home/sschaiter/git/clazzes-osgitools/src/clazzes/osgitools/zephyr_cfg.yml") as fh:
        yaml_content = fh.read()
    conf.loadTestDefFromYaml(yaml_content)
    conf.fetchCycleKeysForNames()

    #conf = ZephyrConfig(access, )
    #testAddingResults(conf)
    #testAddingNewTestPlan(conf, "4.2.0")
    
         
    
    



